from __future__ import annotations

import math
from abc import ABC, abstractmethod
from typing import Dict, Set, Union

import numpy as np
from numpy.typing import NDArray

"""
Defines the pile property geometry classes.

Notes:
- The type of abstract-method property decorators are ignored as a workaround
    to a known mypy issue. (https://github.com/python/mypy/issues/4165)
"""


def _get_depth_keys(d: Union[str, dict], depth: int) -> Set[str]:
    if isinstance(d, str):
        return set([])
    if depth == 0:
        return set(d.keys())
    else:
        # recurse and flatten
        return set(
            [a for diction in d.values() for a in _get_depth_keys(diction, depth - 1)]
        )


# PileProperties inheritance structure:
#  -------------------------------------
#
# - PileProperties
#   - RoundPileProperties
#   - RectPileProperties
#   - TaperedPileProperties


class PileProperties(ABC):
    """
    abstract base class for geometrical pile properties
    """

    _shape: str

    def __init__(
        self,
        pile_type_specification: Dict[str, str],
        alpha_s_sand: float,
        alpha_s_clay: float | str,
        alpha_p: float,
        beta_p: float,
        pile_tip_factor_s: float | int,
        settlement_curve: int,
        elastic_modulus: float | int,
        negative_fr_delta_factor: float | int,
        adhesion: float | int,
        is_low_vibrating: bool,
        is_auger: bool,
        height_base: float,
        name: str | None = None,
    ):
        """
        Parameters
        ----------
        pile_type_specification
            Keys
                pile_type: str
                    One of ["concrete", "steel", "micro", "wood"]
                specification: str
                    One of ["1", "2", "3", "4", "5", "6", "7"]
                installation: str
                    One of ["A", "B", "C", "D", "E", "F", "G"]
            See Also:
                pilecore/lib/pile_types.json
        height_base : float, int
            height of pile base [m].
        alpha_s_sand
            Alpha s factor for coarse layers used in the positive friction calculation [-].
        alpha_s_clay
            Alpha s factor for soft layers used in the positive friction calculation [-].
        alpha_p
            Alpha p factor used in pile tip resistance calculation [-].
        beta_p
            The beta-factor 𝛽_𝑝 to account for the shape of the pile tip according to NEN9997-1 Fig 7.i.
        pile_tip_factor_s
            Factor s [-] used in pile tip resistance calculation as per NEN 9997-1 7.6.2.3 (h).
        settlement_curve
            Settlement lines for figures 7.n and 7.o of NEN-9997-1
        elastic_modulus : float
            Modulus of elasticity of the pile [MPa].
        negative_fr_delta_factor : float
            factor * φ = δ. This parameter is multiplied with phi to get the delta
            parameter used in negative friction calculation according to NEN-9997-1 7.3.2.2 (e).
            Typically values are 1.0 for piles cast in place, and 0.75 for other pile types.
        adhesion
            Adhesion value [kPa], for if the pile shaft has undergone a special treatment.
            Examples:
                - adhesion = 50 kN/m2 for synthetic coating
                - adhesion = 20 kN/m2 for bentonite
                - adhesion = 10 kN/m2 for bitumen coating
            See 7.3.2.2(d) of NEN 9997-1 for examples.
        is_low_vibrating
            Determines wether the pile has an installation type with low vibration.
        is_auger
            Determines wether the pile the pile is an auger pile or not.
        name
            The name of the pile instance.
        """
        self._name = name
        self._alpha_s_clay = alpha_s_clay
        self._adhesion = adhesion
        self._pile_tip_factor_s = pile_tip_factor_s
        self._pile_type_specification = pile_type_specification
        self._pile_type = pile_type_specification["pile_type"]
        self._installation = pile_type_specification["installation"]
        self._specification = pile_type_specification["specification"]

        self._alpha_s_sand = alpha_s_sand
        self._alpha_p = alpha_p
        self._beta_p = beta_p
        self._settlement_curve = settlement_curve
        self._elastic_modulus = elastic_modulus
        self._negative_fr_delta_factor = negative_fr_delta_factor
        self._is_low_vibrating = is_low_vibrating
        self._is_auger = is_auger

        self._height_base = height_base

    @property
    def name(self) -> str | None:
        """
        Pile name (optional)
        """
        if self._name is not None:
            return str(self._name)
        else:
            return None

    @property
    def alpha_s_clay(self) -> str | float:
        """
        Alpha s factor for soft layers used in the positive friction calculation [-].
        """
        if isinstance(self._alpha_s_clay, str):
            return self._alpha_s_clay
        else:
            return float(self._alpha_s_clay)

    @property
    def alpha_s_sand(self) -> float:
        """
        Alpha s factor for coarse layers used in the positive friction calculation [-].
        """
        return float(self._alpha_s_sand)

    @property
    def alpha_p(self) -> float:
        """
        Alpha p factor used in pile tip resistance calculation [-].
        """
        return float(self._alpha_p)

    @property
    def elastic_modulus(self) -> float:
        """
        Modulus of elasticity of the pile [MPa].
        """
        return float(self._elastic_modulus)

    @property
    def negative_fr_delta_factor(self) -> float:
        """
        factor * φ = δ. This parameter will be multiplied with phi to get the delta
        parameter used in negative friction calculation according to NEN-9997-1 7.3.2.2 (e).
        """
        return float(self._negative_fr_delta_factor)

    @property
    def is_low_vibrating(self) -> bool:
        """
        Determines wether the pile has an installation type with low vibration.
        """
        return bool(self._is_low_vibrating)

    @property
    def is_auger(self) -> bool:
        """
        Determines wether the pile the pile is an auger pile or not.
        """
        return bool(self._is_auger)

    @property
    def adhesion(self) -> float:
        """
        Adhesion value [kPa], used if the pile shaft has undergone a special treatment
        """
        return float(self._adhesion)

    @property
    def shape(self) -> str:
        """
        The shape of the pile (round or rect).
        """
        return str(self._shape)

    @property
    def pile_type(self) -> str:
        """
        The pile type of the pile in the configured locale setting.
        """
        return str(self._pile_type)

    @property
    def installation(self) -> str:
        """
        The installation of the pile.
        """
        return str(self._installation)

    @property
    def specification(self) -> str:
        """
        The specification of the pile.
        """
        return str(self._specification)

    @property
    def height_base(self) -> float:
        """
        Height of the pile base [m].
        """
        return float(self._height_base)

    @property
    def settlement_curve(self) -> int:
        """
        The settlement curve number.
        """
        return int(self._settlement_curve)

    @property
    def pile_tip_factor_s(self) -> float:
        """
        Factor s [-], used in pile tip resistance calculation as per NEN 9997-1 7.6.2.3 (h).
        """
        return float(self._pile_tip_factor_s)

    @property
    def pile_type_specification(self) -> Dict[str, str]:
        """
        A Dictionary with default pile-type definition.
        """
        return self._pile_type_specification

    @property
    @abstractmethod
    def circumference_pile_shaft(self) -> float:
        """
        Circumference of the pile shaft [m].
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def circumference_pile_base(self) -> float:
        """
        Circumference of the pile base [m].
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def area_pile_tip(self) -> float:
        """
        Cross-sectional area of the pile tip [m^2].
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def equiv_base_diameter(self) -> float:
        """
        Equivalent base diameter [m].
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def equiv_shaft_diameter(self) -> float:
        """
        Equivalent shaft diameter [m].
        """
        raise NotImplementedError

    @property
    def beta_p(self) -> float:
        """
        The beta-p factor 𝛽_𝑝 to account for the shape of the pile tip according to NEN9997-1 Fig 7.i.
        """
        return float(self._beta_p)

    @abstractmethod
    def get_circum_vs_depth(
        self,
        pile_tip_level: float | int,
        pile_head_level: float | int,
        depth: NDArray[np.floating],
    ) -> NDArray[np.floating]:
        """
        Returns an array with pile circumferences for the depths in the depth parameter.

        Parameters
        ----------
        pile_tip_level
            pile tip level in the depth array.
        pile_head_level
            pile head level in the depth array.
        depth : numpy.ndarray
            Array with depths below surface [m].
        """
        raise NotImplementedError

    @abstractmethod
    def get_area_vs_depth(
        self,
        pile_tip_level: float | int,
        pile_head_level: float | int,
        depth: NDArray[np.floating],
    ) -> NDArray[np.floating]:
        """
        Returns an array with cross-sectional areas for the depths in the depth parameter.

        Parameters
        ----------
        pile_tip_level
            pile tip level in the depth array.
        pile_head_level
            pile head level in the depth array.
        depth : numpy.ndarray
            Array with depths below surface [m].
        """
        raise NotImplementedError


class RoundPileProperties(PileProperties):
    """
    A class with round pile properties
    """

    _shape = "round"

    def __init__(
        self,
        pile_type_specification: Dict[str, str],
        diameter_base: float | int,
        diameter_shaft: float | int,
        height_base: float | int,
        alpha_s_sand: float,
        alpha_s_clay: float | str,
        alpha_p: float,
        beta_p: float,
        pile_tip_factor_s: float,
        settlement_curve: int,
        elastic_modulus: float | int,
        negative_fr_delta_factor: float | int,
        adhesion: float | int,
        is_low_vibrating: bool,
        is_auger: bool,
        name: str | None = None,
    ):
        """
        Parameters
        ----------
        pile_type_specification
            Keys
                pile_type: str
                    One of ["concrete", "steel", "micro", "wood"]
                specification: str
                    One of ["1", "2", "3", "4", "5", "6", "7"]
                installation: str
                    One of ["A", "B", "C", "D", "E", "F", "G"]
            See Also:
                pilecore/lib/pile_types.json
        diameter_base : float, int
            Diameter of pile base [m].
        diameter_shaft : float, int
            Diameter of pile shaft [m].
        height_base : float, int
            height of pile base [m].
        alpha_s_sand
            Alpha s factor for coarse layers used in the positive friction calculation [-].
        alpha_s_clay
            Alpha s factor for soft layers used in the positive friction calculation [-].
        alpha_p
            Alpha p factor used in pile tip resistance calculation [-].
        beta_p
            The beta-factor 𝛽_𝑝 to account for the shape of the pile tip according to NEN9997-1 Fig 7.i.
        pile_tip_factor_s
            Factor s [-] used in pile tip resistance calculation as per NEN 9997-1 7.6.2.3 (h).
        settlement_curve
            Settlement lines for figures 7.n and 7.o of NEN-9997-1
        elastic_modulus : float
            Modulus of elasticity of the pile [MPa].
        negative_fr_delta_factor : float
            factor * φ = δ. This parameter is multiplied with phi to get the delta
            parameter used in negative friction calculation according to NEN-9997-1 7.3.2.2 (e).
            Typically values are 1.0 for piles cast in place, and 0.75 for other pile types.
        adhesion
            Adhesion value [kPa], for if the pile shaft has undergone a special treatment.
            Examples:
                - adhesion = 50 kN/m2 for synthetic coating
                - adhesion = 20 kN/m2 for bentonite
                - adhesion = 10 kN/m2 for bitumen coating

            See 7.3.2.2(d) of NEN 9997-1 for examples.
        is_low_vibrating : bool
            Determines wether the pile has an installation type with low vibration.
        is_auger : bool
            Determines wether the pile the pile is an auger pile or not.
        name
            The name of the pile instance.
        """
        super().__init__(
            pile_type_specification=pile_type_specification,
            pile_tip_factor_s=pile_tip_factor_s,
            alpha_s_sand=alpha_s_sand,
            alpha_s_clay=alpha_s_clay,
            alpha_p=alpha_p,
            beta_p=beta_p,
            elastic_modulus=elastic_modulus,
            adhesion=adhesion,
            settlement_curve=settlement_curve,
            negative_fr_delta_factor=negative_fr_delta_factor,
            is_low_vibrating=is_low_vibrating,
            is_auger=is_auger,
            height_base=height_base,
            name=name,
        )
        self._diameter_shaft = diameter_shaft
        self._diameter_base = diameter_base

    @property
    def diameter_base(self) -> float:
        """
        Diameter of the pile tip/base [m].
        """
        return float(self._diameter_base)

    @property
    def diameter_shaft(self) -> float:
        """
        Diameter of the pile shaft [m].
        """
        return float(self._diameter_shaft)

    @property
    def circumference_pile_shaft(self) -> float:
        """
        Circumference of the pile shaft [m].
        """
        return float(math.pi * self.diameter_shaft)

    @property
    def circumference_pile_base(self) -> float:
        """
        Circumference of the pile tip/base [m].
        """
        return float(math.pi * self.diameter_base)

    @property
    def area_pile_tip(self) -> float:
        """
        Cross-sectional area of the pile tip/base [m].
        """
        return float(0.25 * math.pi * self.diameter_base**2)

    @property
    def area_pile_shaft(self) -> float:
        """
        Cross-sectional area of the pile shaft [m].
        """
        return float(0.25 * math.pi * self.diameter_shaft**2)

    @property
    def equiv_base_diameter(self) -> float:
        """
        Equivalent base diameter [m]. Equal to base diameter for round piles.
        """
        return float(self.diameter_base)

    @property
    def equiv_shaft_diameter(self) -> float:
        """
        Equivalent shaft diameter [m]. Equal to shaft diameter for round piles.
        """
        return float(self.diameter_shaft)

    def get_diameter_vs_depth(
        self,
        pile_tip_level: float | int,
        pile_head_level: float | int,
        depth: NDArray[np.floating],
    ) -> NDArray[np.floating]:
        """
        Returns pile diameters for given input depths.

        Parameters
        ----------
        pile_tip_level : float
            pile tip level in the depth array.
        pile_head_level : float
            pile head level in the depth array.
        depth : np.array
            Array with depths below surface [m].

        Returns
        -------
        np.array
            Array with pile diameters for the depths in the depth parameter.
        """
        diameter_vs_depth = np.zeros_like(depth)
        diameter_vs_depth[
            (depth >= pile_head_level) & (depth < pile_tip_level - self.height_base)
        ] = self.diameter_shaft
        diameter_vs_depth[
            (depth >= pile_tip_level - self.height_base) & (depth <= pile_tip_level)
        ] = self.diameter_base
        return diameter_vs_depth

    def get_circum_vs_depth(
        self,
        pile_tip_level: float | int,
        pile_head_level: float | int,
        depth: NDArray[np.floating],
    ) -> NDArray[np.floating]:
        """
        Returns pile circumferences for given input depths.

        Parameters
        ----------
        pile_tip_level : float
            pile tip level in the depth array.
        pile_head_level : float
            pile head level in the depth array.
        depth : np.array
            Array with depths below surface [m].

        Returns
        -------
        np.array
            Array with pile circumferences for the depths in the depth parameter.
        """
        return (
            self.get_diameter_vs_depth(
                pile_tip_level=pile_tip_level,
                pile_head_level=pile_head_level,
                depth=depth,
            )
            * math.pi
        )

    def get_area_vs_depth(
        self,
        pile_tip_level: float | int,
        pile_head_level: float | int,
        depth: NDArray[np.floating],
    ) -> NDArray[np.floating]:
        """
        Returns cross-sectional areas of the pile for given input depths.

        Parameters
        ----------
        pile_tip_level : float
            pile tip level in the depth array.
        pile_head_level : float
            pile head level in the depth array.
        depth : np.array
            Array with depths below surface [m].

        Returns
        -------
        np.array
            Array with cross-sectional areas for the depths in the depth parameter.
        """
        return (
            self.get_diameter_vs_depth(
                pile_tip_level=pile_tip_level,
                pile_head_level=pile_head_level,
                depth=depth,
            )
            / 2
        ) ** 2 * math.pi


class RectPileProperties(PileProperties):
    """
    A class with rectangular pile properties
    """

    _shape = "rect"

    def __init__(
        self,
        pile_type_specification: Dict[str, str],
        width_base_large: float | int,
        width_base_small: float | int,
        width_shaft_large: float | int,
        width_shaft_small: float | int,
        height_base: float | int,
        alpha_s_sand: float,
        alpha_s_clay: float | str,
        alpha_p: float,
        beta_p: float,
        pile_tip_factor_s: float | int,
        settlement_curve: int,
        elastic_modulus: float | int,
        negative_fr_delta_factor: float | int,
        adhesion: float | int,
        is_low_vibrating: bool,
        is_auger: bool,
        name: str | None = None,
    ):
        """
        Parameters
        ----------
        pile_type_specification
            Keys
                pile_type: str
                    One of ["concrete", "steel", "micro", "wood"]
                specification: str
                    One of ["1", "2", "3", "4", "5", "6", "7"]
                installation: str
                    One of ["A", "B", "C", "D", "E", "F", "G"]
            See Also:
                pilecore/lib/pile_types.json
        width_base_large: Union[float, int]
            Largest dimension of the pile base [m].
        width_base_small: Union[float, int]
            Smallest dimension of the pile base [m].
        width_shaft_large: Union[float, int]
            Largest dimension of the pile shaft [m].
        width_shaft_small: Union[float, int]
            Smallest dimension of the pile shaft [m].
        height_base: Union[float, int]
            Height of pile base [m].
        alpha_s_sand
            Alpha s factor for coarse layers used in the positive friction calculation [-].
        alpha_s_clay
            Alpha s factor for soft layers used in the positive friction calculation [-].
        alpha_p
            Alpha p factor used in pile tip resistance calculation [-].
        beta_p
            The beta-factor 𝛽_𝑝 to account for the shape of the pile tip according to NEN9997-1 Fig 7.i.
        pile_tip_factor_s
            Factor s [-] used in pile tip resistance calculation as per NEN 9997-1 7.6.2.3 (h).
        settlement_curve
            Settlement lines for figures 7.n and 7.o of NEN-9997-1
        elastic_modulus
            Modulus of elasticity of the pile [MPa].
        negative_fr_delta_factor
            factor * φ = δ. This parameter is multiplied with phi to get the delta
            parameter used in negative friction calculation according to NEN-9997-1 7.3.2.2 (e).
            Typically values are 1.0 for piles cast in place, and 0.75 for other pile types.
        adhesion
            Adhesion value [kPa], for if the pile shaft has undergone a special treatment.
            Examples:
                - adhesion = 50 kN/m2 for synthetic coating
                - adhesion = 20 kN/m2 for bentonite
                - adhesion = 10 kN/m2 for bitumen coating

            See 7.3.2.2(d) of NEN 9997-1 for examples.
        is_low_vibrating
            Determines wether the pile has an installation type with low vibration.
        is_auger
            Determines wether the pile the pile is an auger pile or not.
        name
            The name of the pile instance.
        """
        super().__init__(
            pile_type_specification=pile_type_specification,
            alpha_s_sand=alpha_s_sand,
            alpha_s_clay=alpha_s_clay,
            alpha_p=alpha_p,
            beta_p=beta_p,
            pile_tip_factor_s=pile_tip_factor_s,
            settlement_curve=settlement_curve,
            elastic_modulus=elastic_modulus,
            adhesion=adhesion,
            negative_fr_delta_factor=negative_fr_delta_factor,
            is_low_vibrating=is_low_vibrating,
            is_auger=is_auger,
            height_base=height_base,
            name=name,
        )
        self._width_base_large = width_base_large
        self._width_base_small = width_base_small
        self._width_shaft_large = width_shaft_large
        self._width_shaft_small = width_shaft_small

    @property
    def width_base_large(self) -> float:
        """
        Largest dimension of the pile base [m].
        """
        return self._width_base_large

    @property
    def width_base_small(self) -> float:
        """
        Smallest dimension of the pile base [m].
        """
        return self._width_base_small

    @property
    def width_shaft_large(self) -> float:
        """
        Largest dimension of the pile shaft [m].
        """
        return self._width_shaft_large

    @property
    def width_shaft_small(self) -> float:
        """
        Smallest dimension of the pile shaft [m].
        """
        return self._width_shaft_small

    @property
    def circumference_pile_shaft(self) -> float:
        """
        Circumference of the pile shaft [m].
        """
        return float(2 * self.width_shaft_large + 2 * self.width_shaft_small)

    @property
    def circumference_pile_base(self) -> float:
        """
        Circumference of the pile tip/base [m].
        """
        return float(2 * self.width_base_large + 2 * self.width_base_small)

    @property
    def area_pile_tip(self) -> float:
        """
        Cross-sectional area of the pile tip/base [m].
        """
        return float(self.width_base_large * self.width_base_small)

    @property
    def area_pile_shaft(self) -> float:
        """
        Cross-sectional area of the pile shaft [m].
        """
        return float(self.width_shaft_large * self.width_shaft_small)

    @property
    def equiv_base_diameter(self) -> float:
        """
        Equivalent base diameter [m] according to NEN-9997-1+C2_2017
        paragraphs 1.5.2.106a and 7.6.2.3.(10)(e).

        Specifically: returns self.width_base_small
        if self.width_base_large > (1,5 * self.width_base_small)
        """
        a_min = self.width_base_small
        b_max = self.width_base_large
        if b_max > (1.5 * a_min):
            return a_min
        return float(1.13 * a_min * math.sqrt(b_max / a_min))

    @property
    def equiv_shaft_diameter(self) -> float:
        """
        Equivalent base diameter [m] according to NEN-9997-1+C2_2017
        paragraphs 1.5.2.106b and 7.6.2.3.(10)(e).

        Specifically: returns self.width_shaft_small
        if self.width_shaft_large > (1,5 * self.width_shaft_small)
        """
        a_min = self.width_shaft_small
        b_max = self.width_shaft_large
        if b_max > (1.5 * a_min):
            return a_min
        return float(1.13 * a_min * math.sqrt(b_max / a_min))

    def get_circum_vs_depth(
        self,
        pile_tip_level: float | int,
        pile_head_level: float | int,
        depth: NDArray[np.floating],
    ) -> NDArray[np.floating]:
        """
        Returns pile circumferences for given input depths.

        Parameters
        ----------
        pile_tip_level : float
            pile tip level in the depth array.
        pile_head_level : float
            pile head level in the depth array.
        depth : np.array
            Array with depths below surface [m].

        Returns
        -------
        np.array
            Array with pile circumferences for the depths in the depth parameter.
        """
        circum_vs_depth = np.zeros_like(depth)
        circum_vs_depth[
            (depth >= pile_head_level) & (depth < pile_tip_level - self.height_base)
        ] = self.circumference_pile_shaft
        circum_vs_depth[
            (depth >= pile_tip_level - self.height_base) & (depth <= pile_tip_level)
        ] = self.circumference_pile_base
        return circum_vs_depth

    def get_area_vs_depth(
        self,
        pile_tip_level: float | int,
        pile_head_level: float | int,
        depth: NDArray[np.floating],
    ) -> NDArray[np.floating]:
        """
        Returns cross-sectional areas of the pile for given input depths.

        Parameters
        ----------
        pile_tip_level : float
            pile tip level in the depth array.
        pile_head_level : float
            pile head level in the depth array.
        depth : np.array
            Array with depths below surface [m].

        Returns
        -------
        np.array
            Array with cross-sectional areas for the depths in the depth parameter.
        """
        area_vs_depth = np.zeros_like(depth)
        area_vs_depth[
            (depth >= pile_head_level) & (depth <= pile_tip_level - self.height_base)
        ] = self.area_pile_shaft
        area_vs_depth[
            (depth > pile_tip_level - self.height_base) & (depth <= pile_tip_level)
        ] = self.area_pile_tip
        return area_vs_depth


def create_pile_properties_from_api_response(response_dict: dict) -> PileProperties:
    shape = response_dict["type"]
    props = response_dict["props"]

    kwargs = dict(
        height_base=props["height_base"],
        pile_type_specification=props["pile_type_specification"],
        alpha_s_sand=props["alpha_s_sand"],
        alpha_s_clay=props["alpha_s_clay"],
        alpha_p=props["alpha_p"],
        beta_p=props["beta_p"],
        pile_tip_factor_s=props["pile_tip_factor_s"],
        settlement_curve=props["settlement_curve"],
        elastic_modulus=props["elastic_modulus"],
        negative_fr_delta_factor=props["negative_fr_delta_factor"],
        adhesion=props["adhesion"],
        is_low_vibrating=props["is_low_vibrating"],
        is_auger=props["is_auger"],
    )

    if "name" in props.keys():
        kwargs.update(name=props["name"])

    if shape == "round":
        return RoundPileProperties(
            diameter_base=props["diameter_base"],
            diameter_shaft=props["diameter_shaft"],
            **kwargs,
        )
    elif shape == "rect":
        return RectPileProperties(
            width_base_large=props["width_base_large"],
            width_base_small=props["width_base_small"],
            width_shaft_large=props["width_shaft_large"],
            width_shaft_small=props["width_shaft_small"],
            **kwargs,
        )
    else:
        raise ValueError(f"Received unknown pile-type: {shape}.")



import logging
from typing import List

import numpy as np
import pytest
from pygef.cpt import CPTData

from pypilecore.input import (
    create_grouper_payload,
    create_multi_cpt_payload,
    create_pile_properties_payload,
    create_soil_properties_payload,
)
from pypilecore.results import MultiCPTBearingResults
from pypilecore.results.single_cpt_results import (
    CPTResultsTable,
    SingleCPTBearingResults,
)

LOGGER = logging.getLogger(__name__)


def test_create_soil_properties_payload(
    cpt: CPTData, mock_classify_response: dict
) -> None:
    """
    Test creating the soil-properties payload with the `create_soil_properties_payload`
    function.
    """

    # test creating the soil_properties payload
    payload, passover = create_soil_properties_payload(
        cptdata_objects=[cpt],
        classify_tables={cpt.alias: mock_classify_response},
        groundwater_level_nap=-2.5,
        friction_range_strategy="manual",
    )

    assert isinstance(payload, list)
    assert isinstance(passover, dict)


def test_create_pile_properties_payload_green_path() -> None:
    """
    Test creating the soil-properties payload with the `create_soil_properties_payload`
    function.
    """

    def validate_pile_properties_payload(
        payload: dict, shape: str, props: List[str]
    ) -> None:
        # Check that the payload is a dictionary
        assert isinstance(payload, dict), type(payload)

        # Check that the main keys are present
        for key in ["props", "type"]:
            assert key in payload.keys()

        # Check that the pile shape is correct
        assert payload["type"] == shape

        # Check that the props key contains a dictionary
        assert isinstance(payload["props"], dict), type(payload["props"])

        # Check that the expected keys are present under "props"
        for key in ["pile_type_specification", *props]:
            assert key in payload["props"].keys()
            # Delete the key from the dict
            del payload["props"][key]

        # Check that no unexpected keys were present in the dictionary
        assert len(payload["props"]) == 0

    # test green path for round piles with minimal input
    payload = create_pile_properties_payload(
        pile_type="A",
        specification="concrete",
        installation="1",
        pile_shape="round",
        diameter_base=0.5,
    )

    validate_pile_properties_payload(payload, "round", ["diameter_base"])

    # test green path for round piles with extended input
    payload = create_pile_properties_payload(
        pile_type="A",
        specification="concrete",
        installation="1",
        pile_shape="round",
        diameter_base=0.5,
        diameter_shaft=0.4,
        height_base=1,
    )

    validate_pile_properties_payload(
        payload, "round", ["diameter_base", "diameter_shaft", "height_base"]
    )

    # test green path for round piles with extended input
    payload = create_pile_properties_payload(
        pile_type="A",
        specification="concrete",
        installation="1",
        pile_shape="round",
        diameter_base=0.5,
        diameter_shaft=0.4,
        height_base=1.0,
        settlement_curve=1,
        adhesion=50,
        alpha_p=1,
        alpha_s_clay=0.1,
        alpha_s_sand=0.1,
        beta_p=1,
        pile_tip_factor_s=1,
        elastic_modulus=1e3,
        is_auger=True,
        is_low_vibrating=True,
        negative_fr_delta_factor=1.2,
    )

    validate_pile_properties_payload(
        payload,
        "round",
        [
            "diameter_base",
            "diameter_shaft",
            "height_base",
            "settlement_curve",
            "adhesion",
            "alpha_p",
            "alpha_s_clay",
            "alpha_s_sand",
            "beta_p",
            "pile_tip_factor_s",
            "elastic_modulus",
            "is_auger",
            "is_low_vibrating",
            "negative_fr_delta_factor",
        ],
    )

    # test green path for rect piles
    payload = create_pile_properties_payload(
        pile_type="A",
        specification="concrete",
        installation="1",
        pile_shape="rect",
        width_base_large=0.5,
    )

    validate_pile_properties_payload(payload, "rect", ["width_base_large"])

    # test green path for rect piles with extended input
    payload = create_pile_properties_payload(
        pile_type="A",
        specification="concrete",
        installation="1",
        pile_shape="rect",
        width_base_large=0.5,
        width_base_small=0.4,
    )

    validate_pile_properties_payload(
        payload, "rect", ["width_base_large", "width_base_small"]
    )

    # test green path for rect piles with extended input
    payload = create_pile_properties_payload(
        pile_type="A",
        specification="concrete",
        installation="1",
        pile_shape="rect",
        width_base_large=0.5,
        width_base_small=0.4,
        width_shaft_large=0.4,
        width_shaft_small=0.3,
        height_base=1,
    )

    validate_pile_properties_payload(
        payload,
        "rect",
        [
            "width_base_large",
            "width_base_small",
            "width_shaft_large",
            "width_shaft_small",
            "height_base",
        ],
    )

    # test green path for round piles with extended input
    payload = create_pile_properties_payload(
        pile_type="A",
        specification="concrete",
        installation="1",
        pile_shape="rect",
        width_base_large=0.5,
        width_base_small=0.4,
        width_shaft_large=0.4,
        width_shaft_small=0.3,
        height_base=1,
        settlement_curve=1,
        adhesion=50,
        alpha_p=1,
        alpha_s_clay=0.1,
        alpha_s_sand=0.1,
        beta_p=1,
        pile_tip_factor_s=1,
        elastic_modulus=1e3,
        is_auger=True,
        is_low_vibrating=True,
        negative_fr_delta_factor=1.2,
    )

    validate_pile_properties_payload(
        payload,
        "rect",
        [
            "width_base_large",
            "width_base_small",
            "width_shaft_large",
            "width_shaft_small",
            "height_base",
            "settlement_curve",
            "adhesion",
            "alpha_p",
            "alpha_s_clay",
            "alpha_s_sand",
            "beta_p",
            "pile_tip_factor_s",
            "elastic_modulus",
            "is_auger",
            "is_low_vibrating",
            "negative_fr_delta_factor",
        ],
    )

    # Raise ValueError for unrecognized pile_shape
    with pytest.raises(ValueError):
        create_pile_properties_payload(
            pile_type="A",
            specification="concrete",
            installation="1",
            pile_shape="square",
            width_base_large=0.5,
        )

    # Raise ValueError for missing round geometry
    with pytest.raises(ValueError):
        create_pile_properties_payload(
            pile_type="A",
            specification="concrete",
            installation="1",
            pile_shape="round",
        )

    # Raise ValueErrror for missing base height round pile
    with pytest.raises(ValueError):
        create_pile_properties_payload(
            pile_type="A",
            specification="concrete",
            installation="1",
            pile_shape="round",
            diameter_base=0.5,
            diameter_shaft=0.4,
        )

    # Raise ValueErrror for missing base height rect pile
    with pytest.raises(ValueError):
        payload = create_pile_properties_payload(
            pile_type="A",
            specification="concrete",
            installation="1",
            pile_shape="rect",
            width_base_large=0.5,
            width_base_small=0.4,
            width_shaft_large=0.4,
            width_shaft_small=0.3,
        )

    # Raise ValueError when providing rect geometry for round pile shape
    with pytest.raises(ValueError):
        create_pile_properties_payload(
            pile_type="A",
            specification="concrete",
            installation="1",
            pile_shape="round",
            width_base_large=0.5,
        )

    # Raise ValueError for missing rect geometry
    with pytest.raises(ValueError):
        create_pile_properties_payload(
            pile_type="A",
            specification="concrete",
            installation="1",
            pile_shape="rect",
        )

    # Raise ValueError when providing round geometry for rect pile shape
    with pytest.raises(ValueError):
        create_pile_properties_payload(
            pile_type="A",
            specification="concrete",
            installation="1",
            pile_shape="rect",
            diameter_base=0.5,
        )


def test_create_grouper_payload(
    caplog, mock_multi_cpt_bearing_response, mock_results_passover
) -> None:
    cptgroupresults = MultiCPTBearingResults.from_api_response(
        mock_multi_cpt_bearing_response, mock_results_passover
    )

    payload = create_grouper_payload(
        cptgroupresults.cpt_results.cpt_results_dict, pile_load_uls=100
    )

    # check if pile tip levels are sorted
    assert payload["pile_tip_level"] == [1.0, 0.5, 0.0, -0.5]

    # test value error
    single_cpt_results = cptgroupresults.cpt_results.cpt_results_dict["9"]
    x_coord = single_cpt_results.soil_properties.x
    single_cpt_results.soil_properties.__setattr__("_x", None)
    with pytest.raises(ValueError):
        create_grouper_payload(
            cptgroupresults.cpt_results.cpt_results_dict, pile_load_uls=100
        )
    single_cpt_results.soil_properties.__setattr__("_x", x_coord)

    # Test warning for NaN value in single-cpt results table
    cpt_results_dict = dict(
        test_cpt=SingleCPTBearingResults(
            soil_properties=single_cpt_results.soil_properties,
            pile_head_level_nap=single_cpt_results.pile_head_level_nap,
            results_table=CPTResultsTable.from_sequences(
                pile_tip_level_nap=[-10, -11],
                F_nk_cal=[50, 50],
                F_nk_k=[50, 50],
                F_nk_d=[50, 50],
                R_b_cal=[100, np.nan],
                R_b_k=[100, 100],
                R_b_d=[100, 100],
                R_s_cal=[100, 100],
                R_s_k=[100, 100],
                R_s_d=[100, 100],
                R_c_cal=[200, 200],
                R_c_k=[200, 200],
                R_c_d=[200, 200],
                R_c_d_net=[100, 100],
                F_c_k=[0, 0],
                F_c_k_tot=[100, 100],
                negative_friction_range_nap_btm=[-1, -1],
                negative_friction_range_nap_top=[0, 0],
                positive_friction_range_nap_btm=[-8, -8],
                positive_friction_range_nap_top=[-1, -1],
                q_b_max=[100, 100],
                q_s_max_mean=[100, 100],
                qc1=[100, 100],
                qc2=[100, 100],
                qc3=[100, 100],
                s_b=[1, 1],
                s_el=[1, 1],
                k_v_b=[100, 100],
                k_v_1=[100, 100],
            ),
        )
    )
    with caplog.at_level(logging.WARNING):
        create_grouper_payload(cpt_results_dict, pile_load_uls=100)

        assert (
            "CPT test_cpt has NaN values are present in column R_b_cal. " in caplog.text
        )


def test_create_multi_cpt_payload_defaults(
    cpt: CPTData, mock_classify_response: dict
) -> None:
    """
    Test creating the multi_cpt_results payload with the `create_multi_cpt_payload`
    function with minimal arguments.
    """
    payload, _ = create_multi_cpt_payload(
        pile_tip_levels_nap=[0, 1],
        cptdata_objects=[cpt],
        classify_tables={cpt.alias: mock_classify_response},
        groundwater_level_nap=-2.5,
        pile_type="concrete",
        specification="1",
        installation="A",
        pile_shape="round",
        diameter_base=0.3,
    )

    # Check the payload content

    # We don't completely check the soil_properties list here
    list_soil_properties = payload.pop("list_soil_properties")
    assert isinstance(list_soil_properties, list)
    assert len(list_soil_properties) == 1

    assert payload.pop("excavation_param_t") == 1.0
    assert payload.pop("pile_head_level_nap") == "surface"
    assert payload.pop("rel_pile_load") == 0.7
    assert payload.pop("friction_range_strategy") == "lower_bound"
    assert np.isclose(payload.pop("gamma_f_nk"), 1.0)
    assert np.isclose(payload.pop("gamma_r_b"), 1.2)
    assert np.isclose(payload.pop("gamma_r_s"), 1.2)
    pile_properties = payload.pop("pile_properties")
    assert pile_properties.get("type") == "round"
    pile_props = pile_properties.get("props")
    assert pile_props.pop("diameter_base") == 0.3

    # Check the pile-type-specification
    assert pile_props.pop("pile_type_specification") == dict(
        installation="A", pile_type="concrete", specification="1"
    )
    # Check that no extra keys are present in the pile_type_specification
    assert len(pile_props.keys()) == 0

    assert payload.pop("pile_tip_levels_nap") == [0, 1]
    assert np.isclose(payload.pop("soil_load"), 0.0)
    assert payload.pop("use_almere_rules") is False
    assert payload.pop("stiff_construction") is False

    # Check that no extra keys are present in the payload
    assert len(payload.keys()) == 0


def test_create_multi_cpt_payload_extended_round_pile(
    cpt: CPTData, mock_classify_response: dict
) -> None:
    """
    Test creating the multi_cpt_results payload with the `create_multi_cpt_payload`
    function with extended arguments for a round pile.
    """

    payload, _ = create_multi_cpt_payload(
        pile_tip_levels_nap=[0, 1],
        cptdata_objects=[cpt],
        classify_tables={cpt.alias: mock_classify_response},
        groundwater_level_nap=-2.5,
        pile_type="concrete",
        specification="1",
        installation="A",
        pile_shape="round",
        diameter_base=0.3,
        friction_range_strategy="manual",
        stiff_construction=True,
        cpts_group=[cpt.alias],
        fixed_negative_friction_range_nap=(0.0, 5.0),
        fixed_positive_friction_range_nap=(5.0, "ptl"),
        negative_shaft_friction=0.0,
        apply_qc3_reduction=True,
        relative_pile_load=0.8,
        pile_load_sls=0.0,
        soil_load_sls=1.0,
        pile_head_level_nap=5.0,
        excavation_depth_nap=-2.0,
        excavation_param_t=0.5,
        gamma_f_nk=1.2,
        gamma_r_s=1.5,
        gamma_r_b=1.4,
        individual_negative_friction_range_nap={cpt.alias: (0.0, -5.0)},
        individual_positive_friction_range_nap={cpt.alias: (-5.0, "ptl")},
        diameter_shaft=0.25,
        height_base=0.5,
        settlement_curve=2,
        adhesion=0.01,
        alpha_p=0.01,
        alpha_s_clay=0.01,
        alpha_s_sand=0.01,
        beta_p=0.01,
        pile_tip_factor_s=0.01,
        elastic_modulus=0.01,
        is_auger=True,
        is_low_vibrating=True,
        negative_fr_delta_factor=0.01,
    )

    # Check the payload content

    # We don't completely check the soil_properties list here
    assert isinstance(payload.pop("list_soil_properties"), list)

    assert payload.pop("excavation_param_t") == 0.5
    assert payload.pop("pile_head_level_nap") == 5.0
    assert payload.pop("rel_pile_load") == 0.8
    assert payload.pop("friction_range_strategy") == "manual"
    assert np.isclose(payload.pop("gamma_f_nk"), 1.2)
    assert np.isclose(payload.pop("gamma_r_b"), 1.4)
    assert np.isclose(payload.pop("gamma_r_s"), 1.5)
    pile_properties = payload.pop("pile_properties")
    assert pile_properties.get("type") == "round"
    pile_props = pile_properties.get("props")
    assert np.isclose(pile_props.pop("diameter_base"), 0.3)
    assert np.isclose(pile_props.pop("diameter_shaft"), 0.25)
    assert np.isclose(pile_props.pop("height_base"), 0.5)
    assert pile_props.pop("settlement_curve") == 2
    assert np.isclose(pile_props.pop("adhesion"), 0.01)
    assert np.isclose(pile_props.pop("alpha_p"), 0.01)
    assert np.isclose(pile_props.pop("alpha_s_clay"), 0.01)
    assert np.isclose(pile_props.pop("alpha_s_sand"), 0.01)
    assert np.isclose(pile_props.pop("beta_p"), 0.01)
    assert np.isclose(pile_props.pop("pile_tip_factor_s"), 0.01)
    assert np.isclose(pile_props.pop("elastic_modulus"), 0.01)
    assert pile_props.pop("is_auger") is True
    assert pile_props.pop("is_low_vibrating") is True
    assert np.isclose(pile_props.pop("negative_fr_delta_factor"), 0.01)

    # Check the pile-type-specification
    assert pile_props.pop("pile_type_specification") == dict(
        installation="A", pile_type="concrete", specification="1"
    )
    # Check that no extra keys are present in the pile_type_specification
    assert len(pile_props.keys()) == 0

    assert payload.pop("pile_tip_levels_nap") == [0, 1]
    assert np.isclose(payload.pop("soil_load"), 1.0)
    assert payload.pop("use_almere_rules") is False
    assert payload.pop("stiff_construction") is True

    # Check the optional payload keys
    assert payload.pop("cpts_group") == [cpt.alias]
    assert payload.pop("fixed_negative_friction_range_nap") == (0.0, 5.0)
    assert payload.pop("fixed_positive_friction_range_nap") == (5.0, "ptl")
    assert np.isclose(payload.pop("f_nk"), 0.0)
    assert payload.pop("apply_qc3_reduction") is True
    assert np.isclose(payload.pop("pile_load"), 0.0)
    assert np.isclose(payload.pop("excavation_depth_nap"), -2.0)

    # Check that no extra keys are present in the payload
    assert len(payload.keys()) == 0


def test_create_multi_cpt_payload_extended_rect_pile(
    cpt: CPTData, mock_classify_response: dict
) -> None:
    """
    Test creating the multi_cpt_results payload with the `create_multi_cpt_payload`
    function with extended arguments for a rect pile.
    """

    assert isinstance(cpt.alias, str)

    payload, _ = create_multi_cpt_payload(
        pile_tip_levels_nap=[0, 1],
        cptdata_objects=[cpt],
        classify_tables={cpt.alias: mock_classify_response},
        groundwater_level_nap=-2.5,
        pile_type="concrete",
        specification="1",
        installation="A",
        pile_shape="rect",
        width_base_large=0.3,
        friction_range_strategy="manual",
        stiff_construction=True,
        cpts_group=[cpt.alias],
        fixed_negative_friction_range_nap=(0.0, 5.0),
        fixed_positive_friction_range_nap=(5.0, "ptl"),
        negative_shaft_friction=0.0,
        apply_qc3_reduction=True,
        relative_pile_load=0.8,
        pile_load_sls=0.0,
        soil_load_sls=1.0,
        pile_head_level_nap=5.0,
        excavation_depth_nap=-2.0,
        excavation_param_t=0.5,
        gamma_f_nk=1.2,
        gamma_r_s=1.5,
        gamma_r_b=1.4,
        individual_negative_friction_range_nap={cpt.alias: (0.0, -5.0)},
        individual_positive_friction_range_nap={cpt.alias: (-5.0, "ptl")},
        width_base_small=0.25,
        width_shaft_large=0.3,
        width_shaft_small=0.25,
        height_base=0.5,
        settlement_curve=2,
        adhesion=0.01,
        alpha_p=0.01,
        alpha_s_clay=0.01,
        alpha_s_sand=0.01,
        beta_p=0.01,
        pile_tip_factor_s=0.01,
        elastic_modulus=0.01,
        is_auger=True,
        is_low_vibrating=True,
        negative_fr_delta_factor=0.01,
    )

    # Check the payload content

    # We don't completely check the soil_properties list here
    list_soil_properties = payload.pop("list_soil_properties")
    assert isinstance(list_soil_properties, list)
    assert len(list_soil_properties) == 1

    assert payload.pop("excavation_param_t") == 0.5
    assert payload.pop("pile_head_level_nap") == 5.0
    assert payload.pop("rel_pile_load") == 0.8
    assert payload.pop("friction_range_strategy") == "manual"
    assert np.isclose(payload.pop("gamma_f_nk"), 1.2)
    assert np.isclose(payload.pop("gamma_r_b"), 1.4)
    assert np.isclose(payload.pop("gamma_r_s"), 1.5)
    pile_properties = payload.pop("pile_properties")
    assert pile_properties.get("type") == "rect"
    pile_props = pile_properties.get("props")
    assert np.isclose(pile_props.pop("width_base_large"), 0.3)
    assert np.isclose(pile_props.pop("width_base_small"), 0.25)
    assert np.isclose(pile_props.pop("width_shaft_large"), 0.3)
    assert np.isclose(pile_props.pop("width_shaft_small"), 0.25)
    assert np.isclose(pile_props.pop("height_base"), 0.5)
    assert pile_props.pop("settlement_curve") == 2
    assert np.isclose(pile_props.pop("adhesion"), 0.01)
    assert np.isclose(pile_props.pop("alpha_p"), 0.01)
    assert np.isclose(pile_props.pop("alpha_s_clay"), 0.01)
    assert np.isclose(pile_props.pop("alpha_s_sand"), 0.01)
    assert np.isclose(pile_props.pop("beta_p"), 0.01)
    assert np.isclose(pile_props.pop("pile_tip_factor_s"), 0.01)
    assert np.isclose(pile_props.pop("elastic_modulus"), 0.01)
    assert pile_props.pop("is_auger") is True
    assert pile_props.pop("is_low_vibrating") is True
    assert np.isclose(pile_props.pop("negative_fr_delta_factor"), 0.01)

    # Check the pile-type-specification
    assert pile_props.pop("pile_type_specification") == dict(
        installation="A", pile_type="concrete", specification="1"
    )
    # Check that no extra keys are present in the pile_type_specification
    assert len(pile_props.keys()) == 0

    assert payload.pop("pile_tip_levels_nap") == [0, 1]
    assert np.isclose(payload.pop("soil_load"), 1.0)
    assert payload.pop("use_almere_rules") is False
    assert payload.pop("stiff_construction") is True

    # Check the optional payload keys
    assert payload.pop("cpts_group") == [cpt.alias]
    assert payload.pop("fixed_negative_friction_range_nap") == (0.0, 5.0)
    assert payload.pop("fixed_positive_friction_range_nap") == (5.0, "ptl")
    assert np.isclose(payload.pop("f_nk"), 0.0)
    assert payload.pop("apply_qc3_reduction") is True
    assert np.isclose(payload.pop("pile_load"), 0.0)
    assert np.isclose(payload.pop("excavation_depth_nap"), -2.0)

    # Check that no extra keys are present in the payload
    assert len(payload.keys()) == 0


def test_create_multi_cpt_payload_extended_ocr(
    cpt: CPTData, mock_classify_response: dict
) -> None:
    """
    Test creating the multi_cpt_results payload with the `create_multi_cpt_payload`
    function with extended OCR arguments.
    """

    payload, _ = create_multi_cpt_payload(
        pile_tip_levels_nap=[0, 1],
        cptdata_objects=[cpt],
        classify_tables={cpt.alias: mock_classify_response},
        groundwater_level_nap=-2.5,
        pile_type="concrete",
        specification="1",
        installation="A",
        pile_shape="round",
        diameter_base=0.3,
        ocr=2.5,
    )

    # Check the payload content

    list_soil_properties = payload.get("list_soil_properties")
    assert isinstance(list_soil_properties, list)
    assert len(list_soil_properties) == 1
    assert "ocr" in list_soil_properties[0].keys()
    assert np.isclose(list_soil_properties[0]["ocr"], 2.5)

    # Case 2: set OCR with individual_ocr parameter

    payload, _ = create_multi_cpt_payload(
        pile_tip_levels_nap=[0, 1],
        cptdata_objects=[cpt],
        classify_tables={cpt.alias: mock_classify_response},
        groundwater_level_nap=-2.5,
        pile_type="concrete",
        specification="1",
        installation="A",
        pile_shape="round",
        diameter_base=0.3,
        ocr=1.0,
        individual_ocr={cpt.alias: 2.0},
    )

    # Check the payload content

    list_soil_properties = payload.pop("list_soil_properties")
    assert isinstance(list_soil_properties, list)
    assert "ocr" in list_soil_properties[0].keys()
    assert np.isclose(list_soil_properties[0]["ocr"], 2.0)


def test_create_multi_cpt_payload_errors(
    cpt: CPTData, mock_classify_response: dict
) -> None:
    """
    Test creating the multi_cpt_results payload with the `create_multi_cpt_payload`
    with erroneous input.
    """

    with pytest.raises(ValueError):
        create_multi_cpt_payload(
            pile_tip_levels_nap=[0, 1],
            cptdata_objects=[cpt],
            classify_tables={cpt.alias: mock_classify_response},
            groundwater_level_nap=-2.5,
            pile_type="A",
            specification="concrete",
            installation="1",
            pile_shape="round",
        )

    with pytest.raises(ValueError):
        create_multi_cpt_payload(
            pile_tip_levels_nap=[0, 1],
            cptdata_objects=[cpt],
            classify_tables={cpt.alias: mock_classify_response},
            groundwater_level_nap=-2.5,
            pile_type="A",
            specification="concrete",
            installation="1",
            pile_shape="rect",
        )

def test_soil_load_is_always_float(
    cpt: CPTData, mock_classify_response: dict
) -> None:
    """
    Check that a soil-load value `float(0)` is passed even when the soil_load_sls
    argument is None.
    """

    payload, _ = create_multi_cpt_payload(
        pile_tip_levels_nap=[0, 1],
        cptdata_objects=[cpt],
        classify_tables={cpt.alias: mock_classify_response},
        groundwater_level_nap=-2.5,
        pile_type="concrete",
        specification="1",
        installation="A",
        pile_shape="round",
        diameter_base=0.3,
        soil_load_sls=None,
    )

    assert np.isclose(payload.pop("soil_load"), 0.0)


from __future__ import annotations


def create_pile_properties_payload(
    pile_type: str,
    specification: str,
    installation: str,
    pile_shape: str,
    diameter_base: float | None = None,
    diameter_shaft: float | None = None,
    width_base_large: float | None = None,
    width_base_small: float | None = None,
    width_shaft_large: float | None = None,
    width_shaft_small: float | None = None,
    height_base: float | None = None,
    settlement_curve: float | None = None,
    adhesion: float | None = None,
    alpha_p: float | None = None,
    alpha_s_clay: float | None = None,
    alpha_s_sand: float | None = None,
    beta_p: float | None = None,
    pile_tip_factor_s: float | None = None,
    elastic_modulus: float | None = None,
    is_auger: float | None = None,
    is_low_vibrating: float | None = None,
    negative_fr_delta_factor: float | None = None,
) -> dict:
    """
    Creates a dictionary with the `pile_properties` payload content for the PileCore
    endpoints.

    Note
    ----
    The dictionary should be converted to a jsonifyable message before it can be passed
    to a `requests` call directly, for instance with
    `nuclei.client.utils.python_types_to_message()`.

    Parameters
    ----------
    pile_type:
        The equaly named entry in the "pile_type_specification" settings.
        Accepted values are: ["A","B","C""D","E","F","G"]
    specification:
        The equaly named entry in the "pile_type_specification" settings.
        Accepted values are: ["concrete","steel","micro","wood"]
    installation:
        The equaly named entry in the "pile_type_specification" settings.
        Accepted values are: ["1","2","3","4","5","6","7"]
    pile_shape:
        The shape of the pile.
        Accepted values are: ["round", "rect"]
    diameter_base:
        Pile base diameter [m].
        Only relevant if `shape`="round".
    diameter_shaft:
        Pile shaft diameter [m].
        Only relevant if `shape`="round".
    width_base_large:
        Largest dimension of the pile base [m].
        Only relevant if `shape`="rect".
    width_base_small:
        Smallest dimension of the pile base [m].
        Only relevant if `shape`="rect".
    width_shaft_large:
        Largest dimension of the pile shaft [m].
        Only relevant if `shape`="rect".
    width_shaft_small:
        Smallest dimension of the pile shaft [m].
        Only relevant if `shape`="rect".
    height_base:
        Height of pile base [m]. If None, a pile with constant dimension is inferred.
        Cannot be None if diameter_base and diameter_shaft are unequal.
    settlement_curve:
        Settlement lines for figures 7.n and 7.o of NEN-9997-1 As defined in table 7.c
        of NEN-9997-1. The value is inferred from the pile_type_specifications, but can
        be overwritten.
    adhesion:
        Optional adhesion value [kPa], use it if the pile shaft has undergone a special
        treatment. Examples: - adhesion = 50 kN/m2 for synthetic coating - adhesion = 20
        kN/m2 for bentonite - adhesion = 10 kN/m2 for bitumen coating See 7.3.2.2(d) of
        NEN 9997-1 for examples.
    alpha_p:
        Alpha p factor used in pile tip resistance calculation. The value is inferred
        from the pile_type_specifications, but can be overwritten.
    alpha_s_clay:
        Alpha s factor for soft layers used in the positive friction calculation. If
        None the factor is determined as specified in table 7.d of NEN 9997-1.
    alpha_s_sand:
        Alpha s factor for coarse layers used in the positive friction calculation. The
        value is inferred from the pile_type_specifications, but can be overwritten.
    beta_p:
        Factor s used in pile tip resistance calculation as per NEN 9997-1 7.6.2.3 (h).
        The value is inferred from the pile dimension properties, but can be overwritten.
    pile_tip_factor_s:
        Factor s used in pile tip resistance calculation as per NEN 9997-1 7.6.2.3 (h).
        The value is inferred from the pile dimension properties, but can be overwritten.
    elastic_modulus:
        Modulus of elasticity of the pile [Mpa]. The value is inferred from the
        pile_type_specifications, but can be overwritten.
    is_auger:
        Determines weather the pile the pile is an auger pile or not. The value is
        inferred from the pile_type_specifications, but can be overwritten.
    is_low_vibrating:
        Determines weather the pile has an installation type with low vibration. The
        value is inferred from the pile_type_specifications, but can be overwritten.
    negative_fr_delta_factor:
        factor * φ = δ. This parameter will be multiplied with phi to get the delta
        parameter used in negative friction calculation according to NEN-9997-1 7.3.2.2
        (e). Typically values are 1.0 for piles cast in place, and 0.75 for other pile
        types. The value is inferred from the pile_type_specifications, but can be
        overwritten.

    Returns
    -------
    pile_properties:
        The `pile_properties` payload content of the PileCore-API endpoints.

    Raises
    ------
    ValueError:
        - if `pile_shape`=="round" & `diameter_base` is None
        - if `pile_shape`=="rect" & `width_base_large` is None
        - if `pile_shape` not in ["rect", "round"]
    """
    pile_properties: dict = dict(
        props=dict(
            pile_type_specification=dict(
                pile_type=pile_type,
                specification=specification,
                installation=installation,
            )
        ),
        type=pile_shape,
    )

    if pile_shape == "round":
        if diameter_base is None:
            raise ValueError(
                'A value for `diameter_base` is required for `pile_shape`=="round"'
            )

        pile_properties["props"]["diameter_base"] = diameter_base

        if diameter_shaft is not None:
            if height_base is None:
                raise ValueError(
                    "A value for `height_base` is required when the base and shaft dimensions are unequal"
                )

            pile_properties["props"]["diameter_shaft"] = diameter_shaft

    elif pile_shape == "rect":
        if width_base_large is None:
            raise ValueError(
                'A value for `width_base_large` is required for `pile_shape`=="rect"'
            )
        pile_properties["props"]["width_base_large"] = width_base_large

        if width_base_small is not None:
            pile_properties["props"]["width_base_small"] = width_base_small

        if (
            width_shaft_large is not None or width_shaft_small is not None
        ) and height_base is None:
            raise ValueError(
                "A value for `height_base` is required when the base and shaft dimensions are unequal"
            )
        if width_shaft_large is not None:
            pile_properties["props"]["width_shaft_large"] = width_shaft_large

        if width_shaft_small is not None:
            pile_properties["props"]["width_shaft_small"] = width_shaft_small

    else:
        raise ValueError('pile_shape should be one of ["round", "rect"]')

    if height_base is not None:
        pile_properties["props"]["height_base"] = height_base

    if settlement_curve is not None:
        pile_properties["props"]["settlement_curve"] = settlement_curve

    if adhesion is not None:
        pile_properties["props"]["adhesion"] = adhesion

    if alpha_p is not None:
        pile_properties["props"]["alpha_p"] = alpha_p

    if alpha_s_clay is not None:
        pile_properties["props"]["alpha_s_clay"] = alpha_s_clay

    if alpha_s_sand is not None:
        pile_properties["props"]["alpha_s_sand"] = alpha_s_sand

    if beta_p is not None:
        pile_properties["props"]["beta_p"] = beta_p

    if pile_tip_factor_s is not None:
        pile_properties["props"]["pile_tip_factor_s"] = pile_tip_factor_s

    if elastic_modulus is not None:
        pile_properties["props"]["elastic_modulus"] = elastic_modulus

    if is_auger is not None:
        pile_properties["props"]["is_auger"] = is_auger

    if is_low_vibrating is not None:
        pile_properties["props"]["is_low_vibrating"] = is_low_vibrating

    if negative_fr_delta_factor is not None:
        pile_properties["props"]["negative_fr_delta_factor"] = negative_fr_delta_factor

    return pile_properties
